{\rtf1\ansi\ansicpg1252\cocoartf949\cocoasubrtf460
{\fonttbl\f0\fswiss\fcharset0 Helvetica;}
{\colortbl;\red255\green255\blue255;\red36\green79\blue171;}
\paperw11900\paperh16840\margl1440\margr1440\vieww11900\viewh16820\viewkind1
\deftab720
\pard\pardeftab720\ql\qnatural

\f0\fs24 \cf0 ;-------------------------------------------------------------\
;+\
; NAME:\
;		cbfread\
; PURPOSE:\
; \'a0\'a0\'a0\'a0\'a0\'a0Function reads Crystallographic Binary File (CBF) files\
; \'a0\'a0\'a0\'a0\'a0\'a0written by the Pilatus detector control program camserver\
; \'a0\'a0\'a0\'a0\'a0\'a0at the cSAXS Beamline: {\field{\*\fldinst{HYPERLINK "http://sls.web.psi.ch/view.php/beamlines/cs/index.html"}}{\fldrslt \cf2 \ul \ulc2 http://sls.web.psi.ch/view.php/beamlines/cs/index.html}}\
; CATEGORY:\
;		xxx\
; CALLING SEQUENCE:\
;		output = cbfread( 'filename' )\
; INPUTS:\
;		Filename\
; \'a0\'a0\'a0\'a0\'a0\'a0-> Filename with path to the ".cbf"-file\
; KEYWORD PARAMETERS:\
;		---\
; OUTPUTS:\
;		structure of the form\
; \'a0\'a0\'a0\'a0\'a0\'a0-> \{ header: complete header as an array of strings \'a0$\
; \'a0\'a0\'a0\'a0\'a0\'a0\'a0\'a0\'a0\'a0\'a0data: \'a0\'a0data as 2D array of format double \'a0\'a0\'a0\'a0\}\
; COMMON BLOCKS:\
;		---\
; NOTES:\
;		---\
; MODIFICATION HISTORY:\
;		2009-05-13 written by Patrick Huber\
\
; Copyright (C) 2009, Patrick Huber, Laboratory for Neutron Scattering\
; ETHZ & PSI, CH-5232 Villigen PSI, Switzerland\
;\
; This software may be used, copied, or redistributed as long as it is not\
; sold and this copyright notice is reproduced on each copy made. \'a0This\
; routine is provided as is without any express or implied warranties\
; whatsoever.\
;-\
;-------------------------------------------------------------\
\
\
\
FUNCTION cbfread, filename\
\
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\
; ERROR HANDLING\
Catch, TheError\
IF TheError NE 0 THEN BEGIN\
\'a0\'a0Catch, /Cancel\
\'a0\'a0ok = Dialog_Message(['cbfread: ', !Error_State.Msg], /Error)\
\'a0\'a0RETURN, -1\
ENDIF\
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\
\
; 0: no debug information\
; 1: debug information\
; 2: more information\
debug_level = 0;\
\
\
\
; Check positional parameters\
numParams = N_Params()\
IF numParams NE 1 THEN MESSAGE, 'Must supply a filename.'\
\
\
max_header_length = 4096 \'a0; expected maximum length for the text header\
eoh_signature = STRING([ 12B, 26B, 4B, 213B ]) \'a0; end of header signature\
cbf_signature = '###CBF: VERSION' \'a0; CBF file signature\
\
; read all the data\
fdat = READ_BINARY (filename, DATA_TYPE=1 , ENDIAN= 'little' ) ;DATA_TYPE=1 enspricht BYTE\
IF debug_level EQ 1 THEN BEGIN\
	PRINT, 'Read file: '\
	PRINT, filename\
ENDIF\
\
; search for end of header signature within the expected\
; maximum length of a header\
header = STRMID(fdat, 0, MIN([max_header_length, N_ELEMENTS(fdat)]) )\
end_of_header_pos = STRPOS(header, eoh_signature)\
IF end_of_header_pos LT 1 THEN BEGIN\
\'a0\'a0\'a0MESSAGE, 'No header end signature found.'\
ENDIF\
IF debug_level GT 0 THEN BEGIN\
\'a0\'a0\'a0PRINT, 'Header length is', end_of_header_pos\
ENDIF\
\
; check for CBF signature\
IF STRPOS(header, cbf_signature) EQ -1 THEN BEGIN\
\'a0\'a0\'a0MESSAGE, STRJOIN(['CBF signature "', cbf_signature, '" not found in file "', filename])\
end\
\
; Extract the mandatory information for decompression from the header\
; ASSUMPTIONS THEREBY:\
; -> Lineseperator is the combination of "carriage return" and "line feed"\
; -> ":" or "=" is used to separate the Name and the Value of the variables\
header = STRSPLIT(TEMPORARY(header), STRING([13B, 10B]), /EXTRACT)\
IF debug_level GT 1 THEN BEGIN\
	FOR i=0, N_ELEMENTS(header)-1 DO BEGIN\
		PRINT, header[i]\
	ENDFOR\
ENDIF\
\
ind = WHERE(STRPOS(header, 'conversions=') NE -1)\
compr_type = STRTRIM((STRSPLIT( header[ind], ':=', /EXTRACT))[1],2)\
CASE compr_type OF\
	'"x-CBF_BYTE_OFFSET"': compression_type = 1\
	ELSE: MESSAGE, STRJOIN(['Unknown compression type: ', compr_type ])\
ENDCASE\
ind = WHERE(STRPOS(header, 'Content-Transfer-Encoding:') NE -1)\
content_enc = STRTRIM((STRSPLIT( header[ind], ':=', /EXTRACT))[1],2)\
IF content_enc NE 'BINARY' THEN MESSAGE, 'Unknown content encoding: ', content_enc\
ind = WHERE(STRPOS(header, 'X-Binary-Size:') NE -1)\
no_of_bin_bytes = ULONG64((STRSPLIT( header[ind], ':=', /EXTRACT))[1])\
ind = WHERE(STRPOS(header, 'X-Binary-Element-Type:') NE -1)\
el_type = STRTRIM((STRSPLIT( header[ind], ':=', /EXTRACT))[1],2)\
CASE el_type OF\
	'"signed 32-bit integer"': \'a0bytes_per_pixel = 4\
	ELSE: \'a0MESSAGE, STRJOIN(['Unknown element type: ', el_type ])\
ENDCASE\
ind = WHERE(STRPOS(header, 'X-Binary-Element-Byte-Order:') NE -1)\
bin_el_byte_order = STRTRIM((STRSPLIT( header[ind], ':=', /EXTRACT))[1],2)\
IF bin_el_byte_order NE 'LITTLE_ENDIAN' THEN MESSAGE, 'Unknown binary element byte order: ', bin_el_byte_order\
ind = WHERE(STRPOS(header, 'X-Binary-Size-Fastest-Dimension:') NE -1)\
dim1 = ULONG64((STRSPLIT( header[ind], ':=', /EXTRACT))[1])\
ind = WHERE(STRPOS(header, 'X-Binary-Size-Second-Dimension:') NE -1)\
dim2 = ULONG64((STRSPLIT( header[ind], ':=', /EXTRACT))[1])\
\
IF debug_level EQ 1 THEN BEGIN\
	PRINT, 'Compression: ', compr_type\
	PRINT, 'Content-Transfer-Encoding: ', content_enc\
	PRINT, 'X-Binary-Size: ',no_of_bin_bytes\
	PRINT, 'X-Binary-Element-Type: ', el_type\
	PRINT, 'X-Binary-Element-Byte-Order: ', bin_el_byte_order\
	PRINT, 'X-Binary-Size-Fastest-Dimension: ', dim1\
	PRINT, 'X-Binary-Size-Second-Dimension: ', dim2\
ENDIF\
\
\
; Uncompress the binary data\
; initialize return array\
frame = DBLARR( dim1 * dim2, /NOZERO )\
; get binary Data\
dat_in = fdat[end_of_header_pos+STRLEN(eoh_signature):*]\
IF debug_level GT 1 THEN BEGIN\
	print, 'Length of binary array: ', N_ELEMENTS(dat_in)\
ENDIF\
\
; In byte-offset compression the difference to the previous pixel value is\
; stored as eather a byte, or a 16-bit integer or a 32-bit integer,\
; depending on its size.\
; The sizes above one byte are indicated by the escape sequence -1 in the\
; previous data format, i.e, a 32-bit integer is preceded by the sequence\
; 0x80 (hexadecimal for 128 -> too large for a byte)\
; 0x8000 (hexadecimal for 32768 -> too large for a 16-bit integer).\
\
ind_out = 0ULL\
ind_in = 0ULL\
;val_curr = 0.D\
val_diff = 0.D\
WHILE ind_in LE no_of_bin_bytes-1 DO BEGIN\
	val_diff = DOUBLE(dat_in[ind_in])\
	ind_in += 1ULL\
	IF val_diff NE 128.D THEN BEGIN\
		;if not escaped as -128 (0x80=128) use the current byte as\
\'a0\'a0\'a0	;difference, with manual complement to emulate the sign\
\'a0\'a0\'a0	IF val_diff GE 128.D THEN val_diff -= 256.D\
	ENDIF ELSE BEGIN\
		; otherwise check for 16-bit integer value\
		IF ((dat_in[ind_in] NE 0) OR (dat_in[ind_in+1] NE 128)) THEN BEGIN\
			;if not escaped as -32768 (0x8000) use the current 16-bit\
			;integer as difference\
			val_diff = DOUBLE(dat_in[ind_in])+256.D*DOUBLE(dat_in[ind_in+1])\
			;manual complement to emulate the sign\
			IF val_diff GE 32768.D THEN val_diff -= 65536.D\
			ind_in += 2\
		ENDIF ELSE BEGIN\
			ind_in += 2\
			;if everything else failed use the current 32-bit value as difference\
			val_diff = \'a0DOUBLE(dat_in[ind_in]) + $\
						256.D * DOUBLE(dat_in[ind_in+1]) + $\
						65536.D * DOUBLE(dat_in[ind_in+2]) + $\
						16777216.D * DOUBLE(dat_in[ind_in+3])\
			IF val_diff GE 2147483648.D THEN val_diff -= 4294967296.D\
			ind_in += 4\
		ENDELSE\
	ENDELSE\
	;val_curr +=val_diff\
	;frame[ind_out] = val_curr\
	frame[ind_out] = val_diff\
	ind_out += 1\
ENDWHILE\
\
frame = TOTAL(frame, /CUMULATIVE, /PRESERVE_TYPE)\
frame = REFORM(frame, dim1, dim2, /OVERWRITE)\
\
RETURN, \{ header: header, data:frame\}\
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\
Catch, /Cancel}